;;; **********************************************************************
;;; Copyright (c) 1995 Heinrich Taube and Tobias Kunze.  All rights
;;; reserved. Use and copying of this software and preparation of
;;; derivative works based upon this software are permitted and may be
;;; copied as long as no fees or compensation are charged for use, 
;;; copying, or accessing this software and all copies of this software
;;; include this copyright notice.
;;; **********************************************************************

(in-package :stella)

(defconstant +code-font+ '("Courier" 10 :srcor :plain) )
(defconstant +code-font-height+ (font-height  +code-font+))
(defparameter *algo-browser-width* 360)
(defparameter *algotable-size* (xy *algo-browser-width*
                                   (+ (* +code-font-height+ 2)
                                      (* 7 2))))
(defparameter *algotable-position* (xy 5 (+ +border+ 25)))
(defparameter *algo-browser-size*
  (xy *algo-browser-width*
      (+ (xy-y *algotable-position*)
         25 (xy-y *algotable-size*)
         5 
         25 (xy-y *algotable-size*)
         15)))

(defclass algorithm-browser (browser bgc-mixin)
  ((canonical :accessor algorithm-browser-canonical))
  (:default-initargs
    :view-size *algo-browser-size* :color-p t)) ;(xy 380 140)

(defmethod make-browser ((class (eql 'algorithm-browser)) object &rest args)
  (let ((canonical (canonicalize-algorithm object))
        (width (- (xy-x *algo-browser-size*)
                  (* +border+ 1) 
                  15))
        locals inits body make
        label1 label2 menu1 editor1 divider menu2 editor2 browser)
    (setf locals (canonical-algorithm-vars canonical))
    (setf inits (canonical-algorithm-inits canonical))
    (setf body (canonical-algorithm-body canonical))
    (setf make (canonical-algorithm-make canonical))
    
    (multiple-value-setq (menu1 editor1)
      (make-algo-editor t (find-class 'algorithm)
                        make locals (append inits locals) 
                        :view-nick-name ':ieditor
                        :claims (xy width (xy-y *algotable-size*))))
    (setf divider
          (make-instance 'divider 
            :view-nick-name ':divider
            :view-size (xy (xy-x *algo-browser-size*) 5)))
    (multiple-value-setq (menu2 editor2)
      (make-algo-editor nil (find-class 'algorithm)
                        make locals body ;(append locals body)
                        :view-nick-name ':eeditor
                        :claims (xy width (xy-y *algotable-size*))))
    (ccl:set-view-position menu1 *algotable-position*)
    (ccl:set-view-position editor1 
                           (xy +border+ (+ (xy-y (ccl:view-position menu1))
                                           25)))
    
    (ccl:set-view-position divider (xy 0 (+ (xy-y (ccl:view-position editor1)) 
                                            (xy-y (table-claims editor1)))))
    (ccl:set-view-position menu2 (xy +border+
                                     (+ (xy-y (ccl:view-position divider)) 
                                        (xy-y (ccl:view-size divider)))))
    (ccl:set-view-position editor2 
                           (xy +border+ (+ (xy-y (ccl:view-position menu2)) 
                                           25)))
    (let (str)
      (setf str (format nil "Type: ~:(~A~)"
                        (class-name (class-of object))))
      (setf label1 
            (make-instance 'label
              :view-position (xy 8 8)
              :view-size (xy (+ 4 (ccl:string-width str *capella-font*)) 12)
              :title str
              :view-nick-name ':label1))
      (setf str (format nil "Makes: ~:(~A~)"
                        (class-name (slot-value object 'class))))
      (setf label2 
            (make-instance 'label
              :view-position (xy 138 8)
              :view-size (xy (+ 4 (ccl:string-width str *capella-font*)) 12) 
              :title str
              :view-nick-name ':label2)))
    (setf browser
          (apply #'make-instance 'algorithm-browser 
                 :object object
                 :view-subviews
                 (list
                  (make-instance '3d-framed-view
                    :view-nick-name ':control-view
                    :view-position (xy 0 0) 
                    :view-size (xy (xy-x *algo-browser-size*) 
                                   (- (xy-y (ccl:view-position menu1)) 0))
                    :view-subviews
                    (list
                     label1 label2
                     (make-instance 'button
                       :view-position (xy 258 8)
                       :view-size (xy 40 12)
                       :title "Dump"
                       :action #'(lambda (b)
                                   (define-algorithm-action (ccl:view-window b) t))
                       :view-nick-name ':algo-dump)
                     (make-instance 'button
                       :view-position (xy 300 8)
                       :view-size (xy 40 12)
                       :title "Define"
                       :action #'(lambda (b)
                                   (define-algorithm-action (ccl:view-window b) ))
                       :view-nick-name ':algo-define)))
                  menu1 editor1 divider menu2 editor2)
                 args))
    (setf (slot-value browser 'canonical) canonical)
    (set-browser-title browser)
    ;; activate first edit buffer in first table.
    (select-table-editor editor1)
    (update-scrollbar-visibility editor1)
    (update-scrollbar-visibility editor2)
    (set-window-grow browser *algo-browser-size*)
    browser))

(defmethod ccl:set-mini-buffer ((w algorithm-browser) str &rest args)
  (declare (ignore w str args))
  nil)

(defmethod ccl:window-size-parts ((browser algorithm-browser))
  (when (> (length (ccl:view-subviews browser)) 0)
    (let ((size (ccl:view-size browser))
          (divider (ccl:view-named ':divider browser))
          editor)
      (let ((super (ccl:view-named ':control-view browser)))
        ;; center justify "make:" label
        (ccl:set-view-size super (xy (xy-x size) (view-height super)))
        (let ((view (ccl:view-named ':label2 super)))
          (ccl:set-view-position 
           view (xy (- (floor (xy-x size) 2) (floor (view-width view) 2))
                    8)))
        ;; right justify buttons
        (let* ((define (ccl:view-named ':algo-define super))
               (dump (ccl:view-named ':algo-dump super))
               (width (view-width define)))
          (ccl:set-view-position 
           define (xy (- (xy-x size) width 10) 8))
          (ccl:set-view-position 
           dump (xy (- (xy-x size) (* width 2) (* 10 2)) 8))))
      ;; set listing width to window width - 20 (borders) - 15 (hscoller)
      ;; new height = window height - pos+10 (border)+15 (scroller)
      (when divider
        (ccl:set-view-size divider (xy (xy-x size)
                                       (view-height divider)))
        (ccl:invalidate-view divider t))
      (when (setf editor (ccl:view-named ':ieditor browser))
        (let ((s (xy (- (xy-x size) +border+ 15) (view-height editor))))
          (ccl:set-view-size editor s)
          (setf (table-claims editor) s)
          (size-table editor s)
          (update-scrollbar-visibility editor)
          ))
      (when (setf editor (ccl:view-named ':eeditor browser))
        (let ((s (xy (- (xy-x size) +border+ 15)
                     (- (view-height browser)
                        (xy-y (ccl:view-position editor)) 
                        15))))
          (ccl:set-view-size editor s)
          (setf (table-claims editor) s)
          (size-table editor s)
          (update-scrollbar-visibility editor)
          )))))

(defclass algotable (ccl::scroller-mixin table)
  ()
  (:default-initargs :h-scrollp nil
    :draw-scroller-outline nil :table-color *bgc*))

(defun algorithm-browser? (object)
  (typep object 'algorithm-browser))

(defmethod update-window-for-changed-object ((browser algorithm-browser)
                                              object)
  (if (eq (object-view-object browser) object)
    (set-browser-title browser object)))

(defmethod set-browser-title ((browser algorithm-browser) &optional object)
  (ccl:set-window-title
   browser
   (format nil "~A Code" 
           (object-namestring (or object (object-view-object browser))))))

(defstruct (canonical-algorithm (:type list))
  type name make inits vars body)

(defun claim-available (table available)
  (loop with s = 0 for entry in (table-entries table)
        for y = (+ (table-entry-y entry) 7)
        when (table-entry-visible? entry)
        do (incf s y)
        (when (> s available) 
          (return (let ((r (- s y))) (if (= r 0) available r))))
        finally (return s)))

(defmethod ccl::scroll-bar-limits ((view algotable))
  (let ((extent (slot-value view 'extent)))
    (ccl::normal-scroll-bar-limits view (xy-x extent) (xy-y extent))))

(defun make-algo-editor (init? type make vars forms &rest args) 
  ;; create editor view with entries appropriate for the
  ;; specified type of event data.
  ;; in the edit table.
  (let ((data-slots (or (mapcar #'parameter-name (object-parameters make))
                        (list 'rhythm)))
        (algo-slots (mapcar #'parameter-name (object-parameters type)))
        (local-vars (loop for var in vars by #'cddr collect var))
        (entries '())
        table menu)
    ;; create visible entries for each user form. mark each form found
    ;; by nulling it from its set. 
    (loop with tail and count = 0 
          for (place value) on forms by #'cddr
          do
        ;  (format t "~%place ~S vars=~S" place local-vars)
          (push
           (cond ((eq place ':form)
                  (make-algo-entry (intern (format nil "FORM~D" (incf count)))
                                    ':form value nil nil t))
                 ((setf tail (member place data-slots))
                  (make-algo-entry place ':data-slot value nil))
                 ((setf tail (member place algo-slots))
                  (make-algo-entry place ':algo-slot value nil))
                 ((setf tail (member place local-vars))
                  (if init?
                    (make-algo-entry place ':var value nil nil t) 
                    (make-algo-entry place ':var value nil)))
                 (t
                  (format t
                          "Don't know what to do with ~S ~S"  place value)
                  ))
           entries)
          (when tail (setf (first tail) nil tail nil)))
    ;; now all visible entries have been created, add invisble entries for
    ;; each remaining slot in the slot set.
    (dolist (s data-slots)
      (when s (push (make-algo-entry s ':data-slot nil t) entries)))
    (dolist (s algo-slots)
      (when s (push (make-algo-entry s ':algo-slot nil t) entries)))
    (dolist (s local-vars)
      (when s (push (make-algo-entry s ':var nil t nil init?) entries)))
    (setf table
          (apply #'make-instance 'algotable
                 :view-nick-name (if init? :ieditor :eeditor)
                 :value-args '(algobuffer)
                 :label-args (list 'table-label :view-font +code-font+)
                 :edit-x (xy 75 4) :entries (reverse entries)
                 args))
    (setf (ccl:scroll-bar-scroll-size (ccl::v-scroller table))
          +code-font-height+)
    (setf menu 
          (make-entry-menu 
           init? 
           (sort entries 
                 #'(lambda (x y &aux (l '(:data-slot :algo-slot :var :form)))
                     (< (position x l) (position y l)))
                 :key #'table-entry-type)))
    (values menu table)))

(defun make-algo-entry (form type &optional value (hidden t) label user?)
  (make-instance 'table-entry
    :id form 
    :label (if user?
             (make-instance 'algo-user-label
               :title (string-capitalize (string form))
               :view-size (xy 50 (font-height +code-font+)))
             (or label (format nil "~:(~A~)" form)))
    :mode (logior +entry-moveable+ +entry-hideable+ 
                  (if hidden +entry-hidden+ 0))
    :type type
    :editor value))

(defun canonicalize-algorithm (algorithm)
  (let ((external (slot-value algorithm 'external))
        macro type name make inits vars body)
    (setf macro (pop external))
    (setf type (if (or (null macro) (eq macro 'mute))
                 (find-class 'algorithm)
                 (find-class macro)))
    ;; use actual name instead of external form.
    (pop external)
    (setf name (object-name algorithm))
    ;; use internal object class
    (setf make (slot-value algorithm 'class))
    (unless (eq macro 'mute) (pop external))

    (setf inits (pop external))
    (when (and (consp (car external)) (eq (caar external) 'vars))
      (dolist (form (cdr (pop external)))
        (if (consp form)
          (progn (push (first form) vars) (push (second form) vars))
          (progn (push form vars) (push nil vars))))
      (setf vars (nreverse vars)))
    (dolist (form external)
      (if (and (consp form) (member (first form) '(setf setq)))
        (loop for (place value) on (cdr form) by #'cddr
              do (progn (push place body) (push value body)))
        (progn (push ':form body) (push form body))))
    (setf body (nreverse body))
    (make-canonical-algorithm :type type :name name :make make
                              :inits inits :vars vars :body body)))

;;;
;;;
;;;

(defclass algobuffer (ccl:fred-dialog-item)
  ()
  (:default-initargs
  ;  :wrap-p t 
    :view-font +code-font+
    :allow-returns t))

(defmethod form-to-editor ((buffer algobuffer) value)
  (let ((string "")
        (lines 1)
        (*print-case* ':downcase))
    (if (consp value)
      (setf string 
            (let ((*print-right-margin* 55))
              (format nil "~:w" value))
            lines (1+ (count #\return string)))
      (when value
        (setf string (let ((*print-case* ':downcase))
                       (prin1-to-string value)))))
    (ccl:set-dialog-item-text buffer string)
    (ccl:set-view-size 
     buffer (xy 0 (* (font-height (ccl:view-font buffer)) lines)))))

;;;
;;; this is hacked because deleteing from the end of the buffer screws up:
;;; the line count changes when the curser moves from col 1 to 0, so when
;;; the sursor is at col 0 it occupies a "line" on the screen that isnt
;;; accounted for by lines-in-buffer. this function also doesnt seem
;;; to account for empty lines, ie consecutive #\returns
;;;

(defmethod ccl:view-key-event-handler :around ((view algobuffer) char)
  (if (char= char #\newline)
    (progn (expand-action view)
           (call-next-method))
    (if (char= char #\delete)
      (let* ((buf (ccl:fred-buffer view))
             col entry size )
        (setf col (ccl:buffer-column buf))
        (call-next-method)
        ;(setf siz (ccl:buffer-size buf))
        (when (= col 0)
          (unless nil ;(= siz 0)
            (setf size (* (font-height (ccl:view-font view))
                          (max (ccl:lines-in-buffer buf) 1)))
            (setf entry (ccl:view-container view))
            (ccl:set-view-size 
             view (xy (view-width view) size))
            (setf (table-entry-y entry) (+ size 7))
            (redisplay-changed-table (ccl:view-container entry)))))
      (if (char= char #\tab) 
        (progn
          ; (ccl:ed-indent-for-lisp (ccl:fred-buffer view))
          )
        (call-next-method))
      )
    ))

(defun expand-action (buffer &optional shrink?)
  (let ((fonth (font-height (ccl:view-font buffer)))
        (height (view-height buffer)))
    (unless (and shrink? (< (- height fonth) fonth))
      (let ((entry (ccl:view-container buffer)))
        (when shrink? (setf fonth (- fonth)))
        (incf (table-entry-y entry) fonth)
        (ccl:set-view-size 
         buffer (xyinc (ccl:view-size buffer) 0 fonth))
        (redisplay-changed-table (ccl:view-container entry))))))

;;;
;;; entry menu holds entries to install in table.
;;;

(defclass entry-menu (ccl:pop-up-menu menu-action-mixin)
  ()
  (:default-initargs 
    :view-font (italic-face *capella-font*)
    :view-nick-name ':entry-menu
    :auto-update-default nil
    :view-size (xy 75 18)))

(defun make-entry-menu (init? list)
  (let ((items '())
        vars?)
    ;; add title menu item.
    (push (make-instance 'ccl:menu-item 
            :menu-item-title (if init? "Initialize" "Execute")
            :style :italic :disabled t) 
          items)
    ;; add data slots and delimiter
    (loop while (and list (eq (table-entry-type (first list)) ':data-slot))
          do (push (make-instance 'entry-menu-item :entry (pop list))
                   items))
    (when items
      (push (make-instance 'ccl:menu-item 
              :menu-item-title "-" :disabled t) 
            items))
    ;; add algorithm slots
    (when t ; init?
      (loop while (and list (eq (table-entry-type (first list)) ':algo-slot))
            do (push (make-instance 'entry-menu-item :entry (pop list))
                     items))
      (push (make-instance 'ccl:menu-item 
              :menu-item-title "-" :disabled t) 
            items))
    ;; add  local vars followed by the Var item
    ;(unless init?
    ;  (format t "~%List=~S" (mapcar #'(lambda (x) 
    ;                                  (list
    ;                                   (table-entry-id x)
    ;                                   (table-entry-type x)))
    ;                            list)))
    (loop while (and list (eq (table-entry-type (first list)) ':var))
          do (push (make-instance 'entry-menu-item :entry (pop list))
                   items)
          (setf vars? t))
    (unless INIT? (print vars?))
    (if init?
      (push
       (make-instance 'entry-menu-command :command ':var 
                      :menu-item-title "Add Var")
       items)
      (progn
        (when vars? 
          (push (make-instance 'ccl:menu-item 
                  :menu-item-title "-" :disabled t) 
                items))
        (loop while (and list (eq (table-entry-type (first list)) ':form))
              do (push (make-instance 'entry-menu-item :entry (pop list))
                       items))
        (push
         (make-instance 'entry-menu-command :command ':form 
                        :menu-item-title "Add Form")
         items)))
    (make-instance 'entry-menu
      :view-nick-name (if init? ':imenu ':emenu)
      :menu-items (nreverse items))))

(defclass entry-menu-item (ccl:menu-item)
  ((entry :initform ':slot :accessor entry-menu-item-entry :initarg :entry)))

(defclass entry-menu-command (ccl:menu-item)
  ((command :initarg :command :initform nil)))

(defmethod initialize-instance :after ((item entry-menu-item)
                                       &rest ignore)
  (declare (ignore ignore))
  (ccl:set-menu-item-title 
   item (format nil "~:(~A~)"
                (table-entry-id (slot-value item 'entry)))))

(defmethod ccl:menu-update ((menu entry-menu))
  (dolist (i (ccl:menu-items menu 'entry-menu-item))
    (if (table-entry-hidden? (entry-menu-item-entry i))
      (ccl:menu-item-enable i)
      (ccl:menu-item-disable i))))

(defmethod menu-action (menu (item entry-menu-item))
  (let ((table (ccl:find-named-sibling
                menu 
                (if (eq (ccl:view-nick-name menu) ':imenu)
                  ':ieditor ':eeditor)))
        (entry (entry-menu-item-entry item))
        browser)
    (when (table-entry-hidden? entry)
      (setf browser (view-browser menu))
      (unhide-table-entry entry)
      (redisplay-changed-table table)
      (ccl:set-current-key-handler browser (table-entry-editor entry) t)
      ;; if the entry is large than extent, scroll to upper left
      ;; otherwise scroll to bottom+7
      (let ((scroller (ccl::v-scroller table)))
        (ccl:set-scroll-bar-setting scroller 
                                    (xy-y (ccl:view-position entry)))
        (ccl:dialog-item-action scroller))
      (ccl:menu-item-disable item))))

;;;
;;; variables in the init table and forms in the exec table have
;;; renameable labels.
;;;

(defclass algo-user-label (textbuf)
  ()
  (:default-initargs 
    :view-font +code-font+
    :draw-outline nil :margin 0
    :allow-returns t))

(defmethod ccl:view-key-event-handler :around ((view algo-user-label) char)
  (if (char= char #\newline)
    (rename-algo-user-label view)
    (call-next-method)))

(defun rename-algo-user-label (label)
  (let ((entry (ccl:view-container label))
        (text (textbuf-text label)))
    (if (string= text "")
      (ccl:set-dialog-item-text label (string (table-entry-id entry)))
      (let* ((newid (read-from-string text))
             (table (table-entry-table entry))
             (browser (view-browser entry))
             other)
        (if (setf other (table-find newid table :key #'table-entry-id))
          (unless (eq other entry)
            (warn-user "An entry with the name ~:(~A~) already exists."
                       text)
            (ccl:set-dialog-item-text label (string (table-entry-id entry))))
          (let* ((menu1 (ccl:view-named ':imenu  browser))
                 (menu2 (ccl:view-named ':emenu  browser))
                 (oldid (table-entry-id entry))
                 (title (string oldid))
                 item)
            (when (setf item (ccl:find-menu-item menu1 title))
              (ccl:set-menu-item-title item text))
            (when (setf item (ccl:find-menu-item menu2 title))
              (ccl:set-menu-item-title item text))
            (setf (table-entry-id entry) newid)
            (when (eq (table-entry-type entry) ':var)
              (let* ((othertable (ccl:view-named ':eeditor browser))
                     (otherentry (table-find oldid othertable 
                                             :key #'table-entry-id)))
                (when otherentry
                  (rename-table-entry otherentry newid text))))))))))

(defmethod menu-action (menu (item entry-menu-command))
  (let* ((type (slot-value item 'command))
         (table (ccl:find-named-sibling
                 menu 
                 (if (eq (ccl:view-nick-name menu) ':imenu)
                   ':ieditor ':eeditor)))
         (id (gentemp (string type)))
         entry new)
    (setf entry (make-algo-entry id type nil nil nil t))
    (setf new (make-instance 'entry-menu-item :entry entry))
    (ccl:set-menu-item-action-function
     new #'(lambda () (menu-action menu new)))
    (add-table-entry table entry t)
    
    ;; add new item to menu
    (let ((items (cdr (ccl:menu-items menu))))  ; leave title alone
      (apply #'ccl:remove-menu-items menu items)
      (apply #'ccl:add-menu-items menu
             (loop for i in items
                   if (eq i item) collect new and collect i 
                   else collect i)))))

;;;
;;;
;;;

(defun get-initializations (browser)
  (let ((table (ccl:view-named ':ieditor browser))
        (inits '())
        (vars '()))
    (dolist (e (table-entries table))
      (multiple-value-bind (form flag) (editor-to-form e)
        (when (eq flag ':error)
          (warn-user "Initialization for ~A is unbalanced."
                     (item-title (table-entry-label e)))
          (ccl:set-current-key-handler (ccl:view-window table)
                                       (table-entry-editor e) t)
          (return-from get-initializations ':error))
        (if (eq (table-entry-type e) ':var)
          (progn (push (table-entry-id e) vars)
                 (push form vars))
          (when flag
            (push (table-entry-id e) inits)
            (push form inits)))))
    (values (nreverse inits) (nreverse vars))))

(defun get-executions (browser)
  (let ((table (ccl:view-named ':eeditor browser))
        (forms '()))
    (dolist (e (table-entries table))
      (multiple-value-bind (form flag) (editor-to-form e)
        (when (eq flag ':error)
          (warn-user "Execution for ~A is unbalanced."
                     (item-title (table-entry-label e)))
          (ccl:set-current-key-handler (ccl:view-window table)
                                       (table-entry-editor e) t)
          (return-from get-executions ':error))
        (when flag
          (if (eq (table-entry-type e) ':form)
            (push ':form forms)
            (push (table-entry-id e) forms))
          (push form forms))))
    (nreverse forms)))

(defun define-algorithm-action (browser &optional dump?)
  (let ((internal (algorithm-browser-canonical browser))
        inits vars forms external)
    (multiple-value-setq (inits vars) (get-initializations
                                       browser))
    (when (or (eq inits ':error) (eq vars ':error)
              (eq (setf forms (get-executions browser))
                  ':error))
      (return-from define-algorithm-action ':error))
    (setf (canonical-algorithm-inits internal) inits)
    (setf (canonical-algorithm-vars internal) vars)
    (setf (canonical-algorithm-body internal) forms)
    (setf external (canonical-to-external internal))
    (if dump?
      (let ((w (make-instance 'ccl:fred-window
                 :window-show nil
                 :view-size #@(500 200)
                 :window-title 
                 (format nil "~:(~A~) Code Dump" 
                         (canonical-algorithm-name internal)))))
        (format w "~(~:w~)" external)
        (ccl:fred-update w)
        (ccl:window-set-not-modified #-ccl-3 w #+ccl-3 (ccl:fred-item w))
        (ccl:window-select w))
      (ccl:eval-enqueue external))))

(defun canonical-to-external (canonical)
  (let (macro make vars body)
    (if (eq (canonical-algorithm-make canonical) 
            (find-class 'rhythmic-element))
      (setf macro 'mute make nil)
      (setf macro (class-name (canonical-algorithm-type canonical))
            make (class-name (canonical-algorithm-make canonical))))
    (setf vars (loop for (var val) on (canonical-algorithm-vars canonical)
                     by #'cddr
                     if val collect (list var val) into l
                     else collect var into l
                     finally (return (and l (list (cons 'vars l))))))
    (setf body (loop for (form val) on (canonical-algorithm-body canonical)
                     by #'cddr
                     if (eq form ':form) collect val
                     else collect (list 'setf form val)))
    `(,macro ,(canonical-algorithm-name canonical)
             ,@(and make (list make)) 
             ,(canonical-algorithm-inits canonical)
             ,@ vars ,@ body)))
                

;;;
;;; diver view separates the two  tables. slidiing the divider resizes the
;;; tables relative to one another.
;;;

(defclass divider (ccl:view)
  ())

(defmethod ccl:view-draw-contents ((view divider)) 
  (let* ((wid (view-width view))
         (pos (xy (- (floor wid 2) 2) 0)))
    (#_MoveTo :long (xy 0 2))
    (#_lineto :long (xy wid 2))
    (ccl:fill-rect view ccl:*black-pattern* pos (xyinc pos 5 5))))

(defmethod ccl:view-click-event-handler ((view divider) where)
  (let ((browser (ccl:view-window view))
        (where-y (xy-y where))
        (dividerpos (ccl:view-position view))
        delta-y mouse-y editor1 editor2  menu2 ;menu1
        editor1size  editor2size 
        editor2pos menu2pos minimumpos maximumpos)
    ;(setf menu1 (ccl:view-named ':imenu browser))
    (setf menu2 (ccl:view-named ':emenu browser))
    (setf editor1 (ccl:view-named ':ieditor browser))
    (setf editor2 (ccl:view-named ':eeditor browser))
    (setf editor1size (ccl:view-size editor1))
    (setf editor2size (ccl:view-size editor2))
    (setf editor2pos (ccl:view-position editor2))
    (setf menu2pos (ccl:view-position menu2))
    (setf where-y (xy-y (ccl:convert-coordinates where view browser)))
    (setf minimumpos (- (xy-y *algotable-position*) 3)) ; 3 for menuborder
    (setf maximumpos (- (xy-y (ccl:view-size browser)) 15))

    (let ((scroller (ccl::v-scroller editor1)))
      (ccl:set-scroll-bar-setting scroller 0)
      (ccl:dialog-item-action scroller))

    (loop while (ccl:mouse-down-p)
          when (<= minimumpos
                   (setf mouse-y (xy-y (ccl:view-mouse-position browser)))
                   maximumpos)
          do
          (setf delta-y (- mouse-y where-y))
          (ccl:set-view-position view (xyinc dividerpos 0 delta-y)) 
          
          ;; add delta to initializations
          (set-table-claims editor1 (xyinc editor1size 0 delta-y))
         ; (ccl:invalidate-view editor1)          

          ;; add -delta to executions and delta to positions
          (set-table-claims editor2 (xyinc editor2size 0 (- delta-y)))
          (ccl:set-view-position editor2 (xyinc editor2pos 0 delta-y))
          (ccl:set-view-position menu2 (xyinc menu2pos 0 delta-y))
          (ccl:window-update-event-handler browser))
    (redisplay-changed-table editor1)
    (redisplay-changed-table editor2)
    ;(size-table editor1 (ccl:view-size editor1))
    ;(size-table editor2 (ccl:view-size editor2))
    ))

